#!/bin/sh
#
# This script is executed repeatedly until the terminal is powered off.
# It can go from the configuration program to the graphical login over and over
#
# Some things must be done only once after booting. The FIRST_TIME variable 
# is used to know if this is indeed the first time through the script.

export PATH=/bin:/usr/bin:/sbin:/usr/sbin:/usr/X11R6/bin
FIRST_TIME=0

# Get the default xterminal parameters

if [ -f /etc/xterminal ]; then
	source /etc/xterminal
else
	MENU_TIMEOUT=5
fi

# The first time rc.sh is run, /proc is not mounted

if [ ! -f /proc/kcore ] ; then
  FIRST_TIME=1
  echo "Xterminal init"

  # Get /proc and the loopback network setup

  mount -n -t proc none /proc
  /sbin/ifconfig lo 127.0.0.1
  /sbin/route add -net 127.0.0.0 netmask 255.0.0.0 dev lo

  # Starting portmap will eliminate some error messages with 2.2 kernels.
  case `uname -r` in
    2.2.*)
      /sbin/portmap
    ;;
  esac
fi

# The terminal should have a directory named after its ethernet address
# on the server.

BASEDIR=/var/xterminals
NAME=`/bin/getrefdir`
CTRLDIR=$BASEDIR/$NAME
SERVER=`/sbin/getdefserver defserver`

# Mount the configuration directory if it exists

if [ ! -d /etc/X11/sysconfig ]; then
  echo "Checking xterminal setup directory $CTRLDIR"
  mount -n $SERVER:$CTRLDIR /etc/X11
fi

# No configuration directory for this terminal, create one

if [ ! -d /etc/X11/sysconfig ]; then
  echo
  echo "********************************************************"
  echo Missing xterminal setup directory $CTRLDIR
  echo Trying to create one with default values

  # Take as "default" value the directory named "preferred", or the most
  # recent configuration directory, of the one named "default" (supplied
  # with the kit) if there are none others.

  if mount -n $SERVER:$BASEDIR /mnt/xterminals; then
    if [ -d /mnt/xterminals/preferred/. ]; then 
      DEFAULT=preferred/.
    else
      DEFAULT=`ls -t /mnt/xterminals | head -1`
    fi

    if [ -n "$DEFAULT" -a "$DEFAULT" != "$NAME" ]; then
		cp -a /mnt/xterminals/$DEFAULT /mnt/xterminals/$NAME
	else
		mkdir -p /mnt/xterminals/$NAME \
			&& mkdir -p /mnt/xterminals/$NAME/sysconfig \
			&& mkdir -p /mnt/xterminals/$NAME/tmp \
			&& mkdir -p /mnt/xterminals/$NAME/xtermrc.d \
			&& /etc/initdev /mnt/xterminals/$NAME/dev
			echo APPSERVER=defserver >/mnt/xterminals/$NAME/sysconfig/xterminal
			echo FONTSERVER=defserver >>/mnt/xterminals/$NAME/sysconfig/xterminal
    fi

    # Remember that this is a new copied configuration. Do not keep
    # the plug'n play configuration as this is not transferable.
    # The configuration directory is touched to change the date
    # making this the newest configuration directory.

    touch /mnt/xterminals/$NAME
    touch /mnt/xterminals/$NAME/sysconfig/NEW
    rm -f /mnt/xterminals/$NAME/isapnp.conf

    # The terminal name is changed to obtain unique names

    XTERMNAME=`cat /mnt/xterminals/$NAME/sysconfig/TERMNAME | \
        sed -e 's/^\(.*[^0-9]\)\([0-9][0-9]*\)[ ]*$/\1/'`
    XTERMNUMBER=`cat /mnt/xterminals/$NAME/sysconfig/TERMNAME | \
        sed -e 's/^\(.*[^0-9]\)\([0-9][0-9]*\)[ ]*$/\2/'`
    if [ -z "$XTERMNUMBER" ]; then XTERMNUMBER=0; fi
    XTERMNUMBER=`expr $XTERMNUMBER + 1`
    echo "$XTERMNAME$XTERMNUMBER" >/mnt/xterminals/$NAME/sysconfig/TERMNAME

    # Finally mount the configuration directory

    umount -n /mnt/xterminals
    mount -n $SERVER:$CTRLDIR /etc/X11
  fi
fi

# At this point we should have the configuration directory mounted as /etc/X11
# We can mount the /dev directory too if not already mounted.

if [ ! -d /etc/X11/sysconfig ]; then
  echo Make sure directory /var/xterminals is exported read/write
else
  echo "Xterminal setup directory mounted"
  if [ -f /etc/sysconfig/xterminals ] ; then
	source /etc/sysconfig/xterminal
  fi
  if [ ! -e /dev/mouse ]; then
    mount -n $SERVER:$CTRLDIR/dev /dev
  fi
fi

# Ask what to do: login/configure/shell

if [ -d /etc/X11/sysconfig ]; then

  # We asked for the menu, a timeout does not make sense
  if [ "$STARTUP" = "menu" ]; then MENU_TIMEOUT=0; fi

  if [ "${STARTUP}" = "menu" -o "${MENU_TIMEOUT}" -gt 0 ]; then
    # Offer a menu with timeout
    if [ -f /etc/sysconfig/NEW ]; then
      # We are not configured, do not offer login mode
      xtermmenu 1 ${MENU_TIMEOUT}
    else
      # Offer all menu items
      xtermmenu 0 ${MENU_TIMEOUT}
    fi
    CHOICE=$?
    if [ "${CHOICE}" -eq 0 ]; then STARTUP="login";
    elif [ "${CHOICE}" -eq 1 ]; then STARTUP="config";
    elif [ "${CHOICE}" -eq 2 ]; then STARTUP="shell";
    fi
  else
    if [ -f /etc/sysconfig/NEW -a "${STARTUP}" = "login" ]; then 
      STARTUP="config"
    fi
  fi
else
  # Something is wrong, the shell is the only alternative, no menu
  STARTUP="shell"
fi

case ${STARTUP} in
  # We just want a shell
  shell)
    # Clear the screen and start the xterm services
    reset; clear
    if [ ${FIRST_TIME} -ne 0 ]; then xtermservices; fi
    exec /bin/sh
  ;;

  # Go in configuration mode, make the other configurations in BASEDIR
  # available as templates
  config) 
    echo Trying to mount $BASEDIR on $SERVER
    if mount -n $SERVER:$BASEDIR /mnt/xterminals; then
      xtermconfig ${CTRLDIR}
      reset; clear
      umount -n /mnt/xterminals
    else
      echo Make sure directory $BASEDIR is exported read/write
      echo via NFS
    fi

    # Start the services after the configuration is done
    if [ ${FIRST_TIME} -ne 0 ]; then xtermservices; fi
    exit 0
  ;;

  # The real thing, we are going in graphical login mode
  login)
    if [ ${FIRST_TIME} -ne 0 ]; then xtermservices; fi

    exec xgo
  ;;
esac


