#! /usr/bin/python --


"""
usage: %(progname)s [--check-xdm-clients] [--kill-xdm-clients]
                      [--check-display-clients] [--kill-display-clients]
                      [--xdm-name name] [--xdm-login-name name]
                      [--ip-start ip ] [--ip-end ip] [--display display]
                      [--port-start port] [--port-end port] [--delay d]
                    [--check-user-processes] [--kill-user-processes]
                      [--uid-start uid] [--uid-end uid]

Version: %(version)s

This program checks the running tasks for lingering processes to report (check)
and terminate (kill). With the kill options, a SIGTERM signal is sent to
the unwanted processes, followed by a SIGKILL signal 30 seconds later to
the same processes if they still exist.

For display-clients, all processes with the specified display as DISPLAY
environment variable are considered.

For xdm-clients, all the processes with tcp connections to client
hosts with ip adresses between ip-start and ip-end (by default all) or
associated with the specified display, and to ports between port-start
and port-end (by default 6000 for the X protocol) are considered.
In most cases this is less reliable than the simpler display-client options.

For each client host, only processes with start time newer than the latest
xdm (specified by xdm-name, gdm by default) and the latest xdm-login
(specified by xdm-login-name, gdmlogin by default) may be legitimate.
Others must be leftovers from previous sessions, should not be there,
and are reported or terminated. The termination is by sending the TERM
signal, waiting 30 seconds (or as specified with the delay option),
and then sending the KILL signal.

With check-user-processes, all the processes with an uid in a certain range
(default 500 to 50000) are listed. The range bounds may be changed with
uid-start and uid-end. Option kill-user-processes terminates these processes
by sending the TERM signal, followed by the KILL signal after 30 seconds
(or as specified in the delay option).


"""

_version = "1.1"



import os, sys, string, time, getopt, socket, shlex, popen2, re

#
# Convert an IP address from ascii to number
#

def AtoN(adr):
  lst = string.split(adr,'.')
  return ((string.atoi(lst[0]) * 256 + string.atoi(lst[1])),
          (string.atoi(lst[2]) * 256 + string.atoi(lst[3])))

#
# Check if an IP address ip1 is LE or GE than ip2
#

def IPLE(ip1,ip2):
  return ip1[0] < ip2[0] or (ip1[0] == ip2[0] and ip1[1] <= ip2[1]) 


def IPGE(ip1,ip2):
  return ip1[0] > ip2[0] or (ip1[0] == ip2[0] and ip1[1] >= ip2[1]) 


#
# Initialize global variables which may be set by command line options
# with default values.
#

checkXDMClients = 0
killXDMClients = 0
checkDisplayClients = 0
killDisplayClients = 0
xdmName = "gdm"
xdmLoginName = "gdmlogin"
ipStart = AtoN("0.0.0.0")
ipEnd = AtoN("255.255.255.255")
portStart = 6000
portEnd = 6000
delay = 30
checkUserProcesses = 0
killUserProcesses = 0
uidStart = 500
uidEnd = 50000


# A list of clients to terminate which have been sent the TERM signal,
# but to which a KILL reminder would be nice

clientsToKill = []


BadToken = ""


# An undesirable process was found. Report and, if the kill option is active,
# terminate the process.

def ReportProcess(pid,reason,check,kill):
  global clientsToKill

  if check:
    print "Process ", pid, " ", reason

  if kill:
    print "Signal TERM to ",pid
    try:
      os.kill(pid,15)
    except:
      print "Error sending signal (process already died?)"
      
    clientsToKill.append(pid)


# Check if all killed clients indeed vanished

def KillRemainingClients():
  global clientsToKill, delay

  time.sleep(delay)
  for pid in clientsToKill:
    print "Signal KILL to ",pid
    try:
      os.kill(pid,9)
    except:
      print "Error sending signal (process already died?)"


# Insure that a token is what is expected.

def CheckToken(token,target):
  global BadToken

  if not token == target:
    raise BadToken


# Convert the token to a number and raise an exception if the token
# is not a valid number

def GetNumber(token):
  return string.atoi(token)


# Get a xxx.xxx.xxx.xxx IP string and convert it to an integer IP address

def GetIP(token):
  return AtoN(token)


# The token must be a string starting with an alphabetic character

def GetName(token):
  global BadToken

  if not (token[0] >= 'A' and token[0] <= 'z'):
    raise BadToken

  return token


#
# List all processes connected to remote clients by X connections
#

def ListXDMClients():
  global ipStart, ipEnd, portStart, portEnd

  clients = []

  # netstat tells about all the processes with tcp connections
  
  (output, input) = popen2.popen2("netstat -a -n -p --tcp 2>/dev/null")
  input.close()

  # Skip the two header lines of netstat

  output.readline()
  output.readline()

  # Have a line oriented lexer which "." accepted within tokens

  lexer = shlex.shlex(output)
  lexer.whitespace = ' \011' # leave out 012 and 015 to see the end of line
  lexer.wordchars = lexer.wordchars + '.' # have 0.0.0.0 as one token

  while 1:
    
    # A sample line looks like:
    #
    # tcp 0 0 0.0.0.0:1319 0.0.0.0:* LISTEN 2145/cdplayer_apple 
    #
    # All the "." ":" and "/" are separate tokens. Sometimes the
    # process ID / name may not appear and be replaced by a "-"

    # End of line

    token = lexer.get_token()
    if token == "":
      break

    # Check that all the tokens are of the correct type and gather the needed
    # information from each field.

    try:
      localIP = ""
      localPort = ""
      remoteIP = ""
      remotePort = ""
      connectionStatus = ""
      pid = -1

      CheckToken(token,"tcp")
      token = GetNumber(lexer.get_token())  # eat receive queue length
      token = GetNumber(lexer.get_token())  # eat send queue length
      localIP = GetIP(lexer.get_token())
      CheckToken(lexer.get_token(),":")
      localPort = GetNumber(lexer.get_token())
      sremoteIP = lexer.get_token()
      remoteIP = GetIP(sremoteIP)
      CheckToken(lexer.get_token(),":")

      # The remote end may not have a port connected yet, represented as a "*"

      token = lexer.get_token()
      if token == "*":
        remotePort = -1
      else:
        remotePort = GetNumber(token)

      connectionStatus = GetName(lexer.get_token())

      # Sometimes netstat may not find the process number and has a "-"
      token = lexer.get_token()
      if not (token == "-"):
        pid = GetNumber(token)
        spid = token

    except:
      print "Malformed netstat line", localIP, localPort, remoteIP, remotePort, connectionStatus, pid


    # Eat the end of line

    while 1:
      token = lexer.get_token()
      if token == "": break

      if token == "tcp":
        lexer.push_token(token)
        break

    # At this point we have a good entry

    if (pid > 0 and IPGE(remoteIP,ipStart) and IPLE(remoteIP,ipEnd) and 
        remotePort >= portStart and remotePort <= portEnd):
      clients.append([spid,remoteIP])

  return clients


#
# List all clients started by sessions on remote displays.
#

def ListDisplayClients():
  global ipStart, ipEnd

  clients = []

  number = re.compile("^[0-9][0-9]*$")
  
  files = os.listdir("/proc")
  for spid in files:
    if number.search(spid) != None:
      try:
        tmpfile = open("/proc/" + spid + "/environ","r")
        tmplexer = shlex.shlex(tmpfile)
        tmplexer.wordchars = tmplexer.wordchars + '.-+:'
        display = ""
        while 1:
          token = tmplexer.get_token()
          if token == "":
            break

          if token == "DISPLAY":
            CheckToken(tmplexer.get_token(),"=")
            token = tmplexer.get_token()
            display = re.sub('^([^:]*):.*$','\\1',val)
            if display != "":
              sremoteIP = socket.gethostbyname(display)
              remoteIP = GetIP(sremoteIP)
              if (IPGE(remoteIP,ipStart) and IPLE(remoteIP,ipEnd)):
                clients.append([spid,remoteIP])
          else:
            while 1:
              token = tmplexer.get_token()
              if token == "" or token == "\000":
                break

          tmplexer = 0
          tmpfile.close()
          tmpfile = 0
        
      except:
        print "Error reading information about process ", file
        
  return clients


# Find and report all the processes connecting to remote clients and with 
# a start time earlier than the latest X Display Manager (xdm and xdmLogin)
# for that client.

def FindClients():
  global ipStart, ipEnd, portStart, portEnd, xdmName, xdmLoginName
  global checkXDMClients, killXDMClients, clientsToKill

  # Dictionary of clients. The key is the IP number. The value for each client
  # is a list to store the start time of the latest xdm, 
  # of the latest xdmLogin, and of all the relevant processes.

  clients = {}
  clientsToKill = []

  if(checkXDMClients or killXDMClients):
    clientList = ListXDMClients()
  else:
    clientList = ListDisplayClients

  for client in clientList:
    spid = client[0]
    sremoteIP = client[1]

    # Read the program name and start time

    try:
      tmpfile = open("/proc/" + spid + "/stat","r")
      tmplexer = shlex.shlex(tmpfile)
      tmplexer.wordchars = tmplexer.wordchars + '.-+'
      
      token = tmplexer.get_token() # skip pid

      processName = ""
      CheckToken(tmplexer.get_token(),"(")
      while 1:
        token = tmplexer.get_token()
        if token == ")":
          break
        else:
          processName = processName + token
        
      i = 0
      while i < 19:
        token = tmplexer.get_token() # skip 19 other fields
        i = i + 1

      startTime = GetNumber(tmplexer.get_token())
      tmplexer = 0
      tmpfile.close()
      tmpfile = 0

      # Get the processes for that client

      if not clients.has_key(sremoteIP):
        clients[sremoteIP] = [0, 0, {}]

      client = clients[sremoteIP]
      clientProcesses = client[2]

      # Insert the process if not already there

      if not clientProcesses.has_key(spid):
        if processName == xdmName:
          client[0] = max(client[0],startTime)
        elif processName == xdmLoginName:
          client[1] = max(client[1],startTime)

        clientProcesses[spid] = (processName,startTime)

    except:
      print "Unable to check process ",pid," (died?)"
      
  # We now have the list of processes. Find the ones to report

  for client in clients.values():
    for spid in client[2].keys():
      pid = int(spid)
      process = (client[2])[spid]
      if process[1] < client[0]:

        # The process started before the latest xdm!

        ReportProcess(pid,"Process " + process[0] + " started at " +
          str(process[1]) + " before " + xdmName + " " + 
          str(client[0]), checkXDMClients or checkDisplayClients, 
          killXDMClients or killDisplayClients)

      elif process[0] != xdmName and process[1] < client[1]:

        # The process started before the latest gdmlogin!

        ReportProcess(pid,"Process " + process[0] + " started at " +
          str(process[1]) + " before " + xdmLoginName + " " + 
          str(client[1]), checkXDMClients or checkDisplayClients, 
          killXDMClients or killDisplayClients)

  # A reminder that they should have terminated

  if killXDMClients:
    KillRemainingClients()


# Find all user processes

def FindUserProcesses():
  global uidStart, uidEnd, checkUserProcesses, killUserProcesses
  global clientsToKill

  clients = {}
  clientsToKill = []
  number = re.compile("^[0-9][0-9]*$")
  
  files = os.listdir("/proc")
  for file in files:
    if number.search(file) != None:
      try:
        (st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid,
         st_size, st_atime, st_mtime, st_ctime) = os.stat("/proc/" + file)
        if st_uid >= uidStart and st_uid <= uidEnd:

          # Try to get the process name for reporting purposes
        
          processName = ""
          try:
            tmpfile = open("/proc/" + file + "/stat","r")
            tmplexer = shlex.shlex(tmpfile)
            tmplexer.wordchars = tmplexer.wordchars + '.-+'
      
            token = tmplexer.get_token() # skip pid

            CheckToken(tmplexer.get_token(),"(")
            while 1:
              token = tmplexer.get_token()
              if token == ")":
                break
              else:
                processName = processName + token
        
            tmplexer = 0
            tmpfile.close()
            tmpfile = 0
          except:
            print "Unable to get name of process " + file

          # This is a user process
        
          ReportProcess(int(file),processName + " is a user process",
              checkUserProcesses, killUserProcesses)

      except:
        print "Process ", file, "vanished"
        
  if killUserProcesses:
    KillRemainingClients()

    
def usage(progname):
  version = _version
  print __doc__ % vars()

def main(argv, stdout, environ):

  global checkXDMClients, killXDMClients 
  global checkDisplayClients, killDisplayClients 
  global xdmName, xdmLoginName, delay
  global ipStart, ipEnd, portStart, portEnd
  global checkUserProcesses, killUserProcesses, uidStart, uidEnd

  progname = argv[0]
  list, args = getopt.getopt(argv[1:], "", ["help", "version", 
               "check-xdm-clients", "kill-xdm-clients",
               "check-display-clients", "kill-display-clients",
               "xdm-name=", "xdm-login-name=",
               "ip-start=", "ip-end=", "display=",
               "port-start=", "port-end=", "delay=",
               "check-extra", "kill-extra", "baseline=",
               "check-user-processes", "kill-user-processes", "uid-start=",
               "uid-end="])

  if len(args) != 0:
    print progname, ": trailing arguments ", args
    usage(progname)
    return

  for (field, val) in list:
    if field == "--help":
      usage(progname)
      return
    elif field == "--version":
      print progname, _version
      return
    elif field == "--check-xdm-clients":
      checkXDMClients = 1;
    elif field == "--kill-xdm-clients":
      killXDMClients = 1;
    elif field == "--check-display-clients":
      checkDisplayClients = 1;
    elif field == "--kill-display-clients":
      killDisplayClients = 1;
    elif field == "--check-user-processes":
      checkUserProcesses = 1;
    elif field == "--kill-user-processes":
      killUserProcesses = 1;
    elif field == "--xdm-name":
      xdmName = val
    elif field == "--xdm-login-name":
      xdmLoginName = val
    elif field == "--ip-start":
      ipStart = AtoN(val)
    elif field == "--ip-end":
      ipEnd = AtoN(val)
    elif field == "--display":
      display = re.sub('^([^:]*):.*$','\\1',val)
      if display == "":
        return
      display = socket.gethostbyname(display)
      ipStart = AtoN(display)
      ipEnd = ipStart
    elif field == "--port-start":
      portStart = string.atoi(val)
    elif field == "--port-end":
      portEnd = string.atoi(val)
    elif field == "--uid-start":
      uidStart = string.atoi(val)
    elif field == "--uid-end":
      uidEnd = string.atoi(val)
    elif field == "--delay":
      delay = string.atoi(val);
    elif field == "--baseline":
      baseline = val

  if checkXDMClients or killXDMClients or \
      checkDisplayClients or killDisplayClients:
    FindXDMClients()

  if checkUserProcesses or killUserProcesses:
    FindUserProcesses()
    


if __name__ == "__main__":
  main(sys.argv, sys.stdout, os.environ)
