#! /usr/bin/python --


"""
usage: %(progname)s --model model [-u target] [-g group] [--copy-newer] 
                    [--copy-identical] [--keep-old] [--rename-old]
                    [-n][-v][-- target ...]

Version: %(version)s

This program uses user model (by default "zmodel" with home "/home/zmodel"), 
whith home directory /home/model, as a model to clone on all target 
arguments (-u arguments or trailing target user names after a --), 
or on all users with home directories in /home and belonging to group. 
For each file in the model subtree, a copy is created in the target 
user subtree unless one exists with all the same parameters 
(mtime, length, mode). With the --copy-identical
option, all files are copied even when they have the same
parameters. With the --copy-newer argument, files are copied only when
the file in the model is newer than the corresponding file in the
target.

When a copy is made from the model to the target, the existing target file
is deleted. However, if the --keep-old argument is specified, no copy 
takes place if a file with the same name already exists 
in the target. With the --rename-old argument, the existing file is 
renamed by adding ".orig" to it instead of being simply removed. If the 
copied file contains the string "model", it is replaced by the target 
username. Similarly if a symbolic  link value contains the string 
"model" it is also replaced. It is therefore advised to select as 
model a rarely occuring name (such as the default "zmodel"). 
As a special case, the file .netscape/preference.js gets the 
uid argument to its approve_license statement replaced by the 
corresponding target user uid.

Option -v may be used to get a printout of all directories scanned and all
actions performed. Option -n may be used to get a printout of all the
actions that would be done without actually adding or modifying any file.
"""

_version = "1.1"

progname = ""

import sys, os, string, getopt, re, shutil, grp

verbose = 0
action = 1
modelHomePrefix = 0
model = "zmodel"
copyNewer = 0
copyIdentical = 0
keepOld = 0
renameOld = 0
SymbolicLink = "SymbolicLink"
RegularFile = "RegularFile"
NonRegularFile = "NoRegularFile"
Directory = "Directory"


# Messages are sent only in verbose mode

def msg(m):
  global verbose

  if verbose:
    print m


# Dispose of the file target, either by renaming or by removal

def DisposeFile(target):
  global keepOld, renameOld, action

  # The file exists

  if os.path.islink(target) or os.path.exists(target):

    # The existing file will be renamed

    if renameOld:

      # There is an existing backup copy; if it is a directory it should be 
      # removed recursively. A file will be overwritten without problem.

      if os.path.isdir(target + ".orig"):
        msg("rm(" + target + ".orig" + ")")
        if action: ret = Spawnv("/bin/rm",["rm","-r","-f",target + ".orig"])

      # Rename the file

      msg("os.rename(" + target + "," + target + ".orig" + ")")
      if action: os.rename(target,target + ".orig")

    # The existing file will be removed, recursively in the case of a directory

    elif not keepOld:
      if os.path.isdir(target):
        msg("rm -r -f " + target)
        if action: ret = Spawnv("/bin/rm",["rm","-r","-f",target])
      else 
        msg("os.remove(" + target + ")")
        if action: os.remove(target)
    

# Copy all the files from the modelHome to the targetHome
# adjusting if necessary for all occurence of the string "model"
# in files or links of the modelHome

def Clone(model, modelHome, modelList, target, targetHome):
  global keepOld, copyIdentical, copyNewer, Directory, SymbolicLink
  global RegularFile, progname, action

  targetHomeStat = os.stat(targetHome)
  targetUID = targetHomeStat[4]
  targetGID = targetHomeStat[5]

  for modelFile in modelList:
    targetFileName = os.path.join(targetHome,modelFile[0])
    modelFileName = os.path.join(modelHome,modelFile[0])
    if keepOld and os.path.exists(targetFileName):
      i = 0 # Do nothing since we do not want to touch any existing file
    elif modelFile[1] is Directory:
      modelFileStat = modelFile[2]

      if os.path.isdir(targetFileName):

        # The directory already exists in the target, insure that the mode
        # is the same.

        targetFileStat = os.stat(targetFileName)
        if modelFileStat[0] != targetFileStat[0]:
          msg("os.chmod("+targetFileName+",",modelFileStat[0],")")
          if action: os.chmod(targetFileName,modelFileStat[0])

      else:

        # The directory does not exist in the target, create it

        DisposeFile(targetFileName)

        msg("os.mkdir(" + targetFileName + ")")
        if action: os.mkdir(targetFileName)

        msg("os.chmod(" + targetFileName + ",",modelFileStat[0],")")
        if action: os.chmod(targetFileName,modelFileStat[0])

        msg("os.chown(" + targetFileName + ",",targetUID,",",targetGID,")")
        if action: os.chown(targetFileName,targetUID,targetGID)

    elif modelFile[1] is SymbolicLink:

      # It is a symbolic link
      # replace model by target in the link if necessary

      if modelFile[2]:
        targetLink = string.replace(modelFile[3],model,target)
      else:
        targetLink = modelFile[3]

      # If the link already points to the right place, do nothing.
      # Otherwise create a new symlink

      if (os.path.islink(targetFileName) and
         os.readlink(targetFileName) == targetLink):
        i = 0 # Do nothing the link is already correct
      else:
        DisposeFile(targetFileName)
        msg("os.symlink(" + targetLink + "," + targetFileName + ")")
        if action: os.symlink(targetLink,targetFileName)

    elif modelFile[1] is RegularFile:

      # It is a regular file.

      modelFileStat = modelFile[3]
      keepFile = 1;
      if os.path.isfile(targetFileName) and (not copyIdentical):
        targetFileStat = os.stat(targetFileName)
        if copyNewer and (targetFileStat[8] > modelFileStat[8]):
          i = 0 # Do nothing, the target is a file different but 
                # newer than model. We want to keep it
        elif ((targetFileStat[8] == modelFileStat[8]) and
             (targetFileStat[6] == modelFileStat[6])):
          i = 0
	  # The target has the same size and modification time
          # Assume they are identical, no copy needed
        else:
          # The files are not identical and we are allowed to replace it
          keepFile = 0
      else:
        # Either we want to copy in any case (copy identical) or the target
        # is not a file
        keepFile = 0

      if keepFile:
        i = 0 # Keep the file as is, nothing to do
      else:

        # Get rid of the file

        DisposeFile(targetFileName)

        if modelFile[0] == ".netscape/preferences.js":
          specialReplace = 1
        else:
          specialReplace = 0

        if modelFile[2] or specialReplace:

          # Process the file line by line

          inputFile = open(modelFileName,"r")
          msg("outputFile = open(" + targetFileName + ',"w")')
          if action: outputFile = open(targetFileName,"w")
          else: outputFile = open("/dev/null","w")

          while 1:
            next = inputFile.readline()
            if not next:
              break

            # Replace the model UID by the target UID

            if specialReplace:
              next = re.sub('browser.startup.license_accepted", "[0-9][0-9]* ',
                       'browser.startup.license_accepted", "' + 
                       str(targetUID) + ' ',next)

            # We need to replace the model username by the target username

            next = string.replace(next,model,target)

            outputFile.write(string.replace(next,model,target))

          inputFile.close()
          outputFile.close()

          msg("shutil.copystat("+modelFileName+","+targetFileName+")")
          if action: shutil.copystat(modelFileName,targetFileName)
        else:

          # No replacement needed, simply copy the file

          msg("shutil.copy2("+modelFileName+","+targetFileName+")")
          if action: shutil.copy2(modelFileName,targetFileName)

        # Set the ownership to target for the new file

        msg("os.chown("+targetFileName+",",targetUID,",",targetGID,")")
        if action: os.chown(targetFileName,targetUID,targetGID)


def Spawnv(path,args):
  pid = os.fork()
  if pid == 0:
    os.execv(path,args)
  else:
    (pid,code) = os.waitpid(pid,0)
    return code


# Gather the list of files in the modelHome

def WalkDir(arg, dirname, names):
  global Directory, SymbolicLink, RegularFile, NonRegularFile, model
  global modelHomePrefix

  fileList = arg

  for name in names:
    fullName = os.path.join(dirname,name)
    relName = fullName[modelHomePrefix + 1:]

    if verbose:
      print fullName
      print relName

    if os.path.islink(fullName):
      linkValue = os.readlink(fullName)
      linkChange = (string.find(linkValue,model) >= 0)
      fileList.append((relName,SymbolicLink,linkChange,linkValue))
    elif os.path.isdir(fullName):
      fileList.append((relName,Directory,os.lstat(dirname)))
    elif os.path.isfile(fullName):
      changeFile = (Spawnv("/bin/grep",["grep","-q",model,fullName]) == 0)
      fileList.append((relName,RegularFile,changeFile,os.lstat(fullName)))
    else:
      fileList.append((relName,NonRegularFile))
      print progname, ": cannot copy non regular file ", fullName


# List all the files in the modelHome

def ListFiles(dir):
  fileList = []
  os.path.walk(dir,WalkDir,fileList)
  return fileList


def usage(progname):
  version = _version
  print __doc__ % vars()


def main(argv, stdout, environ):
  global copyNewer, copyIdentical, keepOld, renameOld, model, progname
  global SymbolicLink, RegularFile, NonRegularFile, Directory
  global modelHomePrefix, action, verbose

  group = ""

  progname = argv[0]
  list, targetUsers = getopt.getopt(argv[1:], "nvu:g:", ["help", "version", 
               "model=", \
               "copy-newer", "copy-identical", "keep-old", "rename-old"])

  for (field, val) in list:
    if field == "--help":
      usage(progname)
      return
    elif field == "--version":
      print progname, _version
      return
    elif field == "-u":
      targetUsers.append(val);
    elif field == "-n":
      action = 0
      verbose = 1
    elif field == "-v":
      verbose = 1
    elif field == "-g":
      group = val
    elif field == "--model":
      model = val
    elif field == "--copy-newer":
      copyNewer = 1
    elif field == "--copy-newer":
      copyNewer = 1
    elif field == "--copy-identical":
      copyIdentical = 1
    elif field == "--keep-old":
      keepOld = 1
    elif field == "--rename-old":
      renameOld = 1

  modelHome = "/home/" + model
  modelHomePrefix = len(modelHome)
  modelList = ListFiles(modelHome)
  for user in targetUsers:
    userHome = "/home/" + user
    Clone(model, modelHome, modelList, user, userHome)

  if group != "":
    groupID = (grp.getgrnam(group))[2]
    for user in os.listdir("/home"):
      userHome = "/home/" + user
      userStat = os.stat(userHome)
      if userStat[5] == groupID:
        Clone(model, modelHome, modelList, user, userHome)

if __name__ == "__main__":
  main(sys.argv, sys.stdout, os.environ)
